/* $Id: grphfile.cpp,v 1.1.1.1 1995/08/16 15:33:26 cvs Exp $
 *
 * Copyright (c) 1996 University of Alberta
 *
 * May be copied or modified under the terms of the GNU General Public
 * License.  See COPYING file for more information.
 */

#include <owl\owlcore.h>
#include <alloc.h>
#include <fstream.h>
#pragma hdrstop

#include "image.h"

#define ABS(i) ((i)<0 ? -(i) : (i))

//  Structures

struct dateTime {
  short int month, day, year;
  short int hour, minute, second;
  };
struct version {
  short int number;
  char letter;
  };
struct TGAtime {
  short int hour, minute, second;
  };
struct ratio {
  short int numerator, denominator;
  };
struct TGAHeader {
  uint8 	IDLen;
  uint8 	colorMapType;
  uint8		imageType;
  uint16	colorMapIndex;
  uint16	colorMapLen;
  uint8 	colorMapEntrySize;
  uint16	xOrigin;
  uint16	yOrigin;
  uint16	imageWidth;
  uint16	imageHeight;
  uint8 	pixelDepth;
  uint8 	imageDesc;
  };
struct TGAFooter {
  long extensionOffset;
  long developerOffset;
  char signature[18];
  };
struct TGAExtension {
  unsigned short int size;
  char authorName[41];
  char authorComments[4][81];
  dateTime dateTime;
  char jobName[41];
  TGAtime jobTime;
  char softwareID[41];
  version softwareVer;
  long keyColor;
  ratio aspectRatio;
  ratio gammaValue;
  long colorCorrectionOffset;
  long postageStampOffset;
  long scanLineOffset;
  unsigned char attributeType;
  };
struct TGADeveloper {
  int16 empty;
  };

struct PCXPaletteEntry {
  uint8 red, green, blue;
  };
struct PCXHeader {
  uint8 	manufacturer;
  uint8 	version;
  uint8 	encoding;
  uint8 	pixelDepth;
  uint16	xMin;
  uint16	yMin;
  uint16	xMax;
  uint16	yMax;
  uint16	hDPI;
  uint16	vDPI;
  PCXPaletteEntry colorMap[16];
  uint8 	reserved;
  uint8 	numPlanes;
  uint16	lineSize;
  uint16	paletteInfo;
  uint16	width;
  uint16	height;
  uint8 	filler[54];
  };
struct PCXFooter {
  uint8		signature;
  PCXPaletteEntry palette[256];
  };

struct BITMAPARRAYHEADER {
  uint16 baType;
  uint32 baSize;
  uint32 baNext;
  uint16 baScreenWidth;
  uint16 baScreenHeight;
  };


//  Macros

#define MIN(int1, int2) ((int1)<(int2) ? (int1) : (int2))
#define MAX(int1, int2) ((int1)>(int2) ? (int1) : (int2))


//  Constants

#define IMAGE_MINFILESIZE 128

const char TGASignature[] = {"TRUEVISION-XFILE."};
const char TGASoftware[41] = {"TGA support for C++ by Chris Studholme"};
const struct version TGASoftwareVersion = {100, ' '};

//  Global Methods

ImageFile* DetermineImageTypeAndOpen(char *fname, int access)
  {
  fstream* f = new fstream (fname, ios::in | ios::binary);
  if ((!f)||(f->bad())) return 0;

  char *extension=strrchr(fname, '.');
  if (extension) extension++;
  int extensionType = 0;
  int i;
  int found=0;
  ImageFile *result = 0;

  if ((extension)&&(extension[0]))
	{
	if (BMPFile::IsExtensionThisType(extension))
	  {
	  extensionType=1;
	  if ((i=BMPFile::IsFileThisType(f))!=0)
		{ found=1; if (i==1) result = new BMPFile(fname, access); }
	  }
	else if (PCXFile::IsExtensionThisType(extension))
	  {
	  extensionType=2;
	  if ((i=PCXFile::IsFileThisType(f))!=0)
		{ found=1; if (i==1) result = new PCXFile(fname, access); }
	  }
	else if (TGAFile::IsExtensionThisType(extension))
	  {
	  extensionType=3;
	  if ((i=TGAFile::IsFileThisType(f))!=0)
		{ found=1; if (i==1) result = new TGAFile(fname, access); }
	  }
	}

  if (!found)
	{
	if ((extensionType!=1)&&((i=BMPFile::IsFileThisType(f))!=0))
	  { if (i==1) result = new BMPFile(fname, access); }
	else if ((extensionType!=2)&&((i=PCXFile::IsFileThisType(f))!=0))
	  { if (i==1) result = new PCXFile(fname, access); }
	else if ((extensionType!=3)&&((i=TGAFile::IsFileThisType(f))!=0))
	  { if (i==1) result = new TGAFile(fname, access); }
	}

  delete f;
  return result;
  }


//  ImageFile Methods

ImageFile::ImageFile (char *fname, int aFlags)
	: Image(aFlags)
  {
  flags &= FLAGS_PUBLICMASK;
  file = INVALID_HANDLE_VALUE;
  fmapping = 0;
  image = 0;

  int access=0;
  if (flags&IMAGE_READ) access |= GENERIC_READ;
  if (flags&IMAGE_WRITE) access |= GENERIC_WRITE | GENERIC_READ;
  int sharing = (flags&IMAGE_READ) ? FILE_SHARE_READ : 0;
  int openmode = (flags&IMAGE_WRITE) ?
						 ((flags&IMAGE_READ) ? OPEN_ALWAYS : CREATE_ALWAYS)
						 : OPEN_EXISTING;

  int fileflags = FILE_ATTRIBUTE_NORMAL;
  if ((flags&IMAGE_READ)&&(flags&IMAGE_WRITE)) fileflags |= FILE_FLAG_RANDOM_ACCESS;
  else fileflags |= FILE_FLAG_SEQUENTIAL_SCAN;

  file = CreateFile(fname, access, sharing, 0, openmode, fileflags, 0);
  filesize = (file!=INVALID_HANDLE_VALUE ? GetFileSize(file,0) : 0);

  if ((flags&IMAGE_WRITE)&&(filesize<IMAGE_MINFILESIZE)&&(file!=INVALID_HANDLE_VALUE))
	{
	char t[IMAGE_MINFILESIZE];
	memset(&t, 0, IMAGE_MINFILESIZE);
	unsigned long written=0;
	WriteFile(file, &t, IMAGE_MINFILESIZE-filesize, &written, 0);
	}

  int paging = (flags&IMAGE_WRITE) ? PAGE_READWRITE : PAGE_READONLY;
  if (file!=INVALID_HANDLE_VALUE)
	fmapping = CreateFileMapping(file, 0, paging, 0, 0, 0);

  int accessmode = (flags&IMAGE_WRITE) ? FILE_MAP_WRITE : FILE_MAP_READ;

  if (fmapping)
	image = MapViewOfFile(fmapping, accessmode, 0, 0, 0);

  if ((image)&&(flags&IMAGE_WRITE)&&(filesize<IMAGE_MINFILESIZE))
	memset(&((char*)image)[filesize], 0, IMAGE_MINFILESIZE-filesize);

  filesize = (file!=INVALID_HANDLE_VALUE ? GetFileSize(file,0) : 0);
  }

ImageFile::~ImageFile ()
  {
  if (image) UnmapViewOfFile(image);
  if (fmapping) CloseHandle(fmapping);
  if (file!=INVALID_HANDLE_VALUE) CloseHandle(file);
  }

BOOL ImageFile::AdjustFileSize(int newsize)
  {
  if ((!image)||(!fmapping)||(file==INVALID_HANDLE_VALUE)||(!(flags&IMAGE_WRITE))) return 0;

  if (newsize > filesize)
	{
	UnmapViewOfFile(image);  image=0;
	CloseHandle(fmapping);   fmapping=0;

	fmapping = CreateFileMapping(file, 0, PAGE_READWRITE, 0, newsize, 0);
	if (fmapping)
	  image = MapViewOfFile(fmapping, FILE_MAP_WRITE, 0, 0, 0);

	if (image)
	  memset(&((char*)image)[filesize], 0, newsize-filesize);

	filesize = GetFileSize(file,0);
	}

  return (filesize==newsize);
  }


//  TGAFile Methods  ///////////////////////////////////////////////////////////

TGAFile::TGAFile(char *fname, int access)
	: ImageFile(fname, access)
  {
  extensionArea=0;
  header=0;
  developerArea=0;
  bits=0;
  SelectImage();
  }

TGAFile::~TGAFile ()
  {
  }

BOOL TGAFile::SelectImage(int index)
  {
  if ((index)||(!image)) return 0;

  extensionArea=0;
  developerArea=0;
  bits=0;
  newTGA=0;
  header=(TGAHeader*) image;

//  if (header->colorMapType) ReadColorMap();

  if (!header->imageType)		// new file
	{
	setmem (header, sizeof(TGAHeader), 0);
	header->IDLen = 0;
	header->colorMapType = 0;
	header->imageType = 2;
	header->colorMapIndex = 0;
	header->colorMapLen = 0;
	header->colorMapEntrySize = 0;
	header->xOrigin = 0;
	header->yOrigin = 0;
	header->imageWidth = 640;
	header->imageHeight = 480;
	header->pixelDepth = 32;
	header->imageDesc = 0;
	newTGA = 0;
	}
  else {
	TGAFooter *footer = (TGAFooter*)&((uint8*)image)[GetCurrentFileSize()-26];
	newTGA = 1;
	int i;
	for (i=1; i<=18; i++)
	  if (footer->signature[i] != TGASignature[i]) newTGA = 0;
	if ((newTGA)&&(footer->extensionOffset))
	  extensionArea = (TGAExtension*)&((uint8*)image)[footer->extensionOffset];
	if ((newTGA)&&(footer->developerOffset))
	  developerArea = (TGADeveloper*)&((uint8*)image)[footer->developerOffset];

	i = sizeof (TGAHeader) + header->IDLen;
	if (header->colorMapType)
	  i += header->colorMapLen * ((header->colorMapEntrySize + 7) >> 3);
	bits = &((uint8*)image)[i];
	readLinePacket = -1;
	}

  return 1;
  }

void TGAFile::CreateFooter()
  {
  }

int TGAFile::GetLineSize()
  {
  return (header->imageWidth * ((header->pixelDepth+7) >> 3));
  }

void TGAFile::GetImageSize(int& width, int& height)
  {
  width = header->imageWidth;
  height = header->imageHeight;
  }

void TGAFile::GetPixelSize(int& bits, int& planes)
  {
  bits = header->pixelDepth;
  planes = 1;
  }

int TGAFile::GetOrientation()
  {
  return (header->imageDesc&0x20 ? 0 : 1 );
  }

/* 			*** Stuff for reading compressed data ***
	pos := 0;
	PixelSize := (Header.PixelDepth+7) div 8;
	for line := 1 to lines do
	  for i := 0 to Header.ImageWidth-1 do
		begin
		  if (i > 0) or (ReadLinePacket = -1) then
		packet := ReadByte else
		begin
		  packet := ReadLinePacket;
		  ReadLinePacket := -1;
		end;
		  if packet and $80 = $80 then
		begin
		  packet := (packet and $7F);
		  for k := 0 to PixelSize-1 do b[k] := ReadByte;
		  if packet > Header.ImageWidth-1-i then
			begin
			  ReadLinePacket := (packet+i-Header.ImageWidth) or $80;
			  packet := Header.ImageWidth-i-1;
			  ReadBufPos := ReadBufPos - PixelSize;
			  if ReadBufPos < 0 then
			SeekRead (FilePos(f)-ReadBufferSize+ReadBufPos);
			end;
		  for j := 0 to packet do
			for k := 0 to PixelSize-1 do
			  PBuf(Buf)^[pos+j*PixelSize+k] := b[k];
		end else
		begin
		  if packet > Header.ImageWidth-1-i then
			begin
			  ReadLinePacket := (packet+i-Header.ImageWidth);
			  packet := Header.ImageWidth-i-1;
			end;
		  for j := 0 to ((packet+1)*PixelSize)-1 do
			PBuf(Buf)^[pos+j] := ReadByte;
		end;
		  pos := pos + (packet+1) * PixelSize;
		  i := i + packet;
		end;
*/

void TGAFile::ReadLineRGB (int y, unsigned char *buf)
  {
  int x, w, h, pixelSize;

  pixelSize = (header->pixelDepth+7) >> 3;
  GetImageSize (w, h);
  if (!header->imageDesc&0x20) y = h-1-y;
  uint8 *tBuf = &bits[y*GetLineSize()];
  int pos=0, fPos=0;

  for (x=0; x<w; x++)
	{
	buf[pos++] = tBuf[fPos+2];
	buf[pos++] = tBuf[fPos+1];
	buf[pos++] = tBuf[fPos];
	fPos += pixelSize;
	}
  }

int TGAFile::SetImageSize (int width, int height)
  {
  header->imageWidth = width;
  header->imageHeight = height;
  return 1;
  }

void TGAFile::SetPixelSize (int bits, int)
  {
  header->pixelDepth = bits;
  }

void TGAFile::WriteLineRGB (int y, unsigned char *buf)
  {
  int x, w, h, pixelSize;

  pixelSize = (header->pixelDepth+7) >> 3;
  GetImageSize (w, h);
  if (!header->imageDesc&0x20) y = h-1-y;
  uint8 *tBuf = &bits[y*GetLineSize()];
  int pos=0, fPos=0;
  for (x=0; x<w; x++)
	{
	tBuf[fPos+2] = buf[pos++];
	tBuf[fPos+1] = buf[pos++];
	tBuf[fPos]   = buf[pos++];
	fPos += pixelSize;
	}
  }

void TGAFile::WriteHeader()
  {
//  WriteColorMap();
  int i = sizeof (TGAHeader) + header->IDLen;
  if (header->colorMapType)
	i += header->colorMapLen * ((header->colorMapEntrySize + 7) >> 3);
  bits = &((uint8*)image)[i];
  }

int TGAFile::IsFileThisType(fstream *file)
  {
  TGAHeader h;
  file->read ((unsigned char*) &h, sizeof(3));  // only 3 bytes needed
  file->clear();
  file->seekg(0);

  if (h.colorMapType&0xFE) return 0;
  switch (h.imageType) {
	case 1:
	case 2:
	case 9:
	case 10:
	  return 1;		// supported type
	case 3:
	case 11:
	case 32:
	case 33:
	  return -1;	// unsupported type
	}

  return 0;		// not Targa image
  }


//  PCXFile Methods  ///////////////////////////////////////////////////////////

PCXFile::PCXFile(char *fname, int access)
	: ImageFile(fname, access)
  {
  header=0;
  footer=0;
  SelectImage();
  }

PCXFile::~PCXFile ()
  {
//  if (linesWritten != -1)
	{
	unsigned char *buf;

//	while (linesWritten < header->yMax-header->yMin+1) Write (buf, 1);
//	if (footer) f->write ((unsigned char*)footer, sizeof (TGAFooter));
	}
  if (footer) delete footer;
  }

BOOL PCXFile::SelectImage(int index)
  {
  if ((index)||(!image)) return 0;

  header=(PCXHeader*) image;
  footer=0;


  if ((!header->manufacturer)&&(!header->version))		// new file
	{
	setmem (&header, sizeof(PCXHeader), 0);
	header->manufacturer = 10;
	header->version = 4;
	header->encoding = 1;
	header->pixelDepth= 8;
	header->xMin = 0;
	header->yMin = 0;
	header->xMax = 639;
	header->yMax = 479;
	header->hDPI = 300;
	header->vDPI = 300;
	header->numPlanes = 1;
	header->lineSize = 640;
	header->paletteInfo = 1;
	header->width = 640;
	header->height = 480;
	imageType = CLUT;
	}
  else {
	if (header->version == 6) imageType = TRUECOLOR;
	if (header->version <= 5) imageType = CLUT;
	if (header->version == 99)
	  {
	  footer = (PCXFooter*)&((uint8*)image)[GetCurrentFileSize()-769];
	  if (footer->signature != 12) footer=0;
	  }
	}

  return 1;
  }

void PCXFile::CreateFooter()
  {
  setmem (footer, sizeof (PCXFooter), 0);
  footer->signature = 12;
  header->version = 5;
  int i;
  for (i=0; i<256; i++)
	footer->palette[i].red=footer->palette[i].green=footer->palette[i].blue = i;
  }

int PCXFile::GetLineSize()
  {
  return (header->numPlanes * header->lineSize);
  }

void PCXFile::GetImageSize (int& width, int& height)
  {
  width  = header->xMax - header->xMin + 1;
  height = header->yMax - header->yMin + 1;
  }

void PCXFile::GetPixelSize (int& bits, int& planes)
  {
  bits = header->pixelDepth;
  planes = header->numPlanes;
  }

void PCXFile::Read (unsigned char *buf, int lines)
  {
  int l, x, i, j;
  unsigned char b;

  for (l = 0; l < lines; l++)
	for (x = 0; x < header->numPlanes * header->lineSize; x++)
	  {
//	  f->get (b);
	  if ((b & 0xC0) == 0xC0)
	{
	  j = b & 0x3F;  if (j == 0) j = 0x40;
//	  f->get (b);
	  for (i=0; i <= j-2; i++)
		buf[l*header->numPlanes*header->lineSize+x+i] = b;
	  x += j-1;
	}
	  buf[l*header->numPlanes*header->lineSize+x] = b;
	  }
  }

void PCXFile::ReadLineRGB (int y, unsigned char *buf)
  {
  int x, w, i=0;
  unsigned char *tBuf = (unsigned char*) malloc (GetLineSize());

  GetImageSize (w, x);
  Read (tBuf, 1);
  switch (imageType) {
	case (TRUECOLOR) :
	  for (x=0; x<w; x++)
		{
		buf[i++] = tBuf[x];
		buf[i++] = tBuf[x+w];
		buf[i++] = tBuf[x+w+w];
		}
	  break;
	case (CLUT):
	case (GRAYSCALE):
	  for (x=0; x<w; x++)
		{
		buf[i++] = tBuf[x];
		buf[i++] = tBuf[x];
		buf[i++] = tBuf[x];
		}
	  break;
	}
  free (tBuf);
  }

int  PCXFile::SetImageSize (int width, int height)
  {
  header->xMin = 0;  header->xMax = width - 1;
  header->yMin = 0;  header->yMax = height - 1;
  header->lineSize = ((width * header->pixelDepth + 15) >> 4) * 2;
  header->width = width;
  header->height = height;
  return 1;
  }

void PCXFile::SetPixelSize (int bits, int planes)
  {
  header->pixelDepth = bits;
  header->numPlanes = planes;
  header->lineSize = (((header->xMax-header->xMin+1) * bits + 15) >> 4)*2;
  }

void PCXFile::GetImageType (int &type)
  {
  type = imageType;
  }

void PCXFile::SetImageType (int type)
  {
  imageType = type;
  switch (type) {
	case (TRUECOLOR) :
	  SetPixelSize (8, 3);
		header->version = 6;
	  break;
	case (GRAYSCALE) :
	  header->version = 5;
	  SetPixelSize (8, 1);
	  if (!footer) CreateFooter ();
	  break;
	case (CLUT) :
	  header->version = 5;
	  SetPixelSize (8, 1);
	  if (!footer) CreateFooter ();
	  break;
	}
  }

void PCXFile::Write (unsigned char *buf, int lines)
  {
  int l, x, i;
  unsigned char b;

//  if (!linesWritten) WriteHeader ();
  for (l=0; l<lines; l++)
	for (x = 0; x < header->numPlanes*header->lineSize; x++)
	  {
	  b = buf[l*header->numPlanes*header->lineSize+x];
	  if ((b == buf[l*header->numPlanes*header->lineSize+x+1]) &&
	  (x < header->numPlanes*header->lineSize-1))
	{
	i = 1;
	while ((b == buf[l*header->numPlanes*header->lineSize+x+i])
	  && (x+i <= header->numPlanes*header->lineSize-1) && (i < 63))
	  i++;
	x += i-1;
//	f->put ((unsigned char) (i+0xC0));
	}
//	else if ((b & 0xC0) == 0xC0) f->put ((unsigned char) 0xC1);
//		f->put (b);
	}
//  linesWritten = linesWritten + lines;
  }

void PCXFile::WriteLineRGB (int y, unsigned char *buf)
  {
  int x, w, i=0;
  unsigned char *tBuf = (unsigned char*) malloc (GetLineSize());

  GetImageSize (w, x);
  switch (imageType) {
	case (TRUECOLOR) :
	  for (x=0; x<w; x++)
		{
		tBuf[x]   = buf[x*3];
		tBuf[w+x] = buf[x*3+1];
		if (header->numPlanes > 2)
		  tBuf[w+w+x] = buf[x*3+2];
		}
	  break;
	case (GRAYSCALE) :
	  for (x=0; x<w; x++)
		tBuf[x] = (buf[i++]+buf[i++]+buf[i++]) / 3;
	  break;
	}
  Write (tBuf, 1);
  free (tBuf);
  }

void PCXFile::WriteHeader()
  {

  }

int PCXFile::IsFileThisType(fstream *file)
  {
  PCXHeader h;
  file->read ((unsigned char*) &h, sizeof(4));  // only 3 bytes needed
  file->clear();
  file->seekg(0);

  if (h.manufacturer!=10) return 0;
  switch (h.version) {
	case 5:
	case 99:
	  return 1;		// supported type
	case 0:
	case 1:
	case 2:
	case 3:
	case 4:
	  return -1;	// unsupported type
	}
  return 0;		// not PCX image
  }


//  BMPFile Methods

BMPFile::BMPFile (char *fname, int aFlags)
	: ImageFile(fname, aFlags)
  {
  BMFH = 0;
  BMIH = 0;
  BMAH = 0;
  bits = 0;
  numImages=0;
  SelectImage(0);
  }

BOOL BMPFile::AdjustFileSize(int newsize)
  {
  if (!image) return 0;

  int offBMAH = BMAH ? ((int)BMAH-(int)image) : -1;	BMAH=0;
  int offBMFH = BMFH ? ((int)BMFH-(int)image) : -1;	BMFH=0;
  int offBMIH = BMIH ? ((int)BMIH-(int)image) : -1;	BMIH=0;
  int offBits = bits ? ((int)bits-(int)image) : -1;	bits=0;

  if (ImageFile::AdjustFileSize(newsize))
	{
	if (offBMAH!=-1) BMAH = (BITMAPARRAYHEADER*)((int)image+offBMAH);
	if (offBMFH!=-1) BMFH = (BITMAPFILEHEADER*)((int)image+offBMFH);
	if (offBMIH!=-1) BMIH = (BITMAPINFOHEADER*)((int)image+offBMIH);
	if (offBits!=-1) bits = (uint8*)((int)image+offBits);

	return 1;
	}

  return 0;
  }

BOOL BMPFile::SelectImage(int index)
  {
  if (!image) return 0;
  if (((uint16*)image)[0]==0x4142)  //'BA'
	flags |= IMAGE_MULTIIMAGE;

  BMAH = 0;
  bits = 0;

  if (flags&IMAGE_MULTIIMAGE) BMAH = (BITMAPARRAYHEADER*)image;
  if ((flags&IMAGE_MULTIIMAGE)&&(BMAH->baType==0x4142)&&(!numImages))
	CountImages();

  if (!(flags&IMAGE_MULTIIMAGE)) numImages=1;
  if (index>=numImages) return 0;

  while ((index)&&(BMAH)&&(BMAH->baType==0x4142)&&(BMAH->baNext))
	BMAH = (BITMAPARRAYHEADER*)&(((uint8*)image)[BMAH->baNext]), index--;

  BMFH = (BITMAPFILEHEADER*) (BMAH ? &(BMAH[1]) : image);
  BMIH = (BITMAPINFOHEADER*) &(BMFH[1]);

  if (BMFH->bfType==0x4D42)			// single image bitmap
	{
	ReadColorMap();
	bits = &((uint8*)image)[BMFH->bfOffBits];
	if (BMIH->biCompression==BI_BITFIELDS)		// BitFields Compression
	  {
	  uint32* mask = (uint32*)(&(((uint8*)BMIH)[BMIH->biSize]));
	  redMask = mask[0];
	  greenMask = mask[1];
	  blueMask = mask[2];
	  }
	}

  if ((BMAH)&&(BMAH->baType!=0x4142)&&(flags&IMAGE_WRITE))
	{
	setmem (BMAH, sizeof(BITMAPARRAYHEADER), 0);
	BMAH->baType	= 0x4142;			//'BA'
	BMAH->baSize	= 1294;			// arbitrary value
	}

  if ((BMFH->bfType!=0x4D42)&&(flags&IMAGE_WRITE))
	{
	setmem (BMFH, sizeof(BITMAPFILEHEADER), 0);
	setmem (BMIH, sizeof(BITMAPINFOHEADER), 0);
	BMFH->bfType    = 0x4D42;       //'BM'
	BMFH->bfSize    = 1294;         //file size in DWORDs
	BMFH->bfOffBits = 1078;         //header size + 1024 color map
	BMIH->biSize    = sizeof(BITMAPINFOHEADER);	//size of this record in bytes
	BMIH->biWidth   = 64;           //width of image in pixels
	BMIH->biHeight  = 64;           //height of image in pixels
	BMIH->biPlanes  = 1;            //must be 1
	BMIH->biBitCount = 8;        //bit depth of each pixel
	BMIH->biSizeImage	= 64*64;	//size of bitmap
	}

  return 1;
  }

BOOL BMPFile::CreateNewImageAndSelect()
  {
  if ((!(flags&IMAGE_MULTIIMAGE))||(!(flags&IMAGE_WRITE))||(!BMAH)) return 0;

  while ((BMAH->baType==0x4142)&&(BMAH->baNext))
	BMAH = (BITMAPARRAYHEADER*)&(((uint8*)image)[BMAH->baNext]);

  if (BMAH->baType!=0x4142) return 0;

  int currentsize = BMFH->bfOffBits + ABS(BMIH->biHeight)*GetLineSize();
  AdjustFileSize(currentsize+sizeof(BMAH)+sizeof(BMFH)+sizeof(BMIH));
  BMAH->baNext = currentsize;
                                                         
  numImages++;
  SelectImage(numImages-1);

  return 1;
  }

void BMPFile::CountImages()
  {
  if (!(flags&IMAGE_MULTIIMAGE)) return;

  numImages=1;
  BITMAPARRAYHEADER *BMAH = (BITMAPARRAYHEADER*)image;
  while ((BMAH->baType==0x4142)&&(BMAH->baNext))
	BMAH = (BITMAPARRAYHEADER*)&(((uint8*)image)[BMAH->baNext]), numImages++;
  }

void BMPFile::ReadColorMap()
  {
  if (!BMIH) return;
  if ((BMIH->biBitCount > 8)&&(BMIH->biClrUsed == 0)) return;

  unsigned int numColors = BMIH->biClrUsed;
  if (!numColors) numColors = 1 << BMIH->biBitCount;
  if (numColors > 256) numColors = 256;
  char *bmimem = new char[sizeof(BITMAPINFOHEADER)+numColors*sizeof(RGBQUAD)];
  if (!bmimem) return;
  BITMAPINFO *bmi = (BITMAPINFO*) bmimem;
  memset(&(bmi->bmiHeader), 0, sizeof(BITMAPINFOHEADER));
  memcpy(&(bmi->bmiHeader), BMIH, MIN(sizeof(BITMAPINFOHEADER),BMIH->biSize));
  bmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
  memcpy(&(bmi->bmiColors), &(((uint8*)BMIH)[BMIH->biSize]), numColors*sizeof(RGBQUAD));

  if (palette) delete palette;
  palette = new TPalette(bmi);
  delete bmimem;
  }

inline int BMPFile::GetLineSize()
  {
  return (BMIH->biWidth*BMIH->biPlanes*BMIH->biBitCount) >> 3;
  }

void BMPFile::GetImageSize (int& width, int& height)
  {
  width = BMIH->biWidth;
  height = ABS(BMIH->biHeight);
  }

void BMPFile::GetPixelSize (int& bits, int& planes)
  {
  bits = BMIH->biBitCount;
  planes = BMIH->biPlanes;
  }

// file types supported:	24 and 32 bit true color images
//							16 and 32 bit bitfields compressed
//                      	4 and 8 bit colormap images
void BMPFile::ReadLineRGB (int y, unsigned char *buf)
  {
  if ((!bits)||(!(flags&IMAGE_READ))) return;

  int x, w, h, pixelSize;
  PALETTEENTRY pe;

  pixelSize = (BMIH->biBitCount+7) >> 3;
  GetImageSize (w, h);
  if (BMIH->biHeight>=0) y = h-1-y;
  uint8 *tBuf = &bits[y*GetLineSize()];

  if ((BMIH->biBitCount==4) && (palette))
	for (x=0; x<w; x++)
	  {
	  int i = tBuf[x/2];
	  palette->GetPaletteEntry(x&1 ? i&0x0F : i>>4, pe);
	  buf[x*3] = pe.peRed;  buf[x*3+1] = pe.peGreen;  buf[x*3+2] = pe.peBlue;
	  }

  else if ((BMIH->biBitCount==8) && (palette))
	for (x=0; x<w; x++)
	  {
	  palette->GetPaletteEntry(tBuf[x], pe);
	  buf[x*3] = pe.peRed;  buf[x*3+1] = pe.peGreen;  buf[x*3+2] = pe.peBlue;
	  }

  else if ((BMIH->biCompression==BI_BITFIELDS)&&((BMIH->biBitCount==16)||(BMIH->biBitCount==32)))
	{
	int bm=blueMask, b=0, gm=greenMask, g=0, rm=redMask, r=0;
	while (bm) bm>>=1, b++;
	while (gm) gm>>=1, g++;
	while (rm) rm>>=1, r++;
	for (x=0; x<w; x++)
	  {
	  uint32 p = (pixelSize==4 ? ((uint32*)tBuf)[x] :
								 ((uint16*)tBuf)[x]);
	  buf[x*3]   = r>8 ? (p&redMask) >> (r-8)   :(p&redMask) << (8-r);
	  buf[x*3+1] = g>8 ? (p&greenMask) >> (g-8) :(p&greenMask) << (8-g);
	  buf[x*3+2] = b>8 ? (p&blueMask) >> (b-8)  :(p&blueMask) << (8-b);
	  }
	}

  else if (BMIH->biBitCount>=24)
	for (x=0; x<w; x++)
	  {
	  buf[x*3] = tBuf[x*pixelSize+2];
	  buf[x*3+1] = tBuf[x*pixelSize+1];
	  buf[x*3+2] = tBuf[x*pixelSize];
	  }
  }

int  BMPFile::SetImageSize (int width, int height)
  {
  if ((!(flags&IMAGE_WRITE))||(bits)||(!BMIH)) return 0;
  BMIH->biWidth = width;
  BMIH->biHeight = height;
  return 1;
  }

void BMPFile::SetPixelSize (int bitCount, int planes)
  {
  if ((!(flags&IMAGE_WRITE))||(bits)||(!BMIH)) return;
  BMIH->biBitCount = bitCount;
  BMIH->biPlanes = planes;
  }

int BMPFile::GetOrientation()
  {
  if (!BMIH) return 1;
  return (BMIH->biHeight<0 ? 0 : 1);
  }

int BMPFile::SetSpecialFileType(long type)
  {
  if ((!BMFH)||(!(flags&IMAGE_WRITE))) return 0;

  BMFH->bfReserved1 = type&0x0FFFF;
  BMFH->bfReserved2 = type>>16;
  return 1;
  }

long BMPFile::GetSpecialFileType()
  {
  if (!BMFH) return 0;
  return BMFH->bfReserved2*0x10000 + BMFH->bfReserved1;
  }

int BMPFile::SetCompression(int compressionType, int redBits, int greenBits, int blueBits)
  {
  if (!BMIH) return 0;
  if (compressionType==IMAGE_NOCOMPRESSION)
	{
	BMIH->biCompression = 0;
	return 1;
	}
  if (((compressionType==BMP_COMPRESSION_BITFIELDS)||(compressionType==IMAGE_DEFAULTCOMPRESSION))&&
	  ((BMIH->biBitCount==16)||(BMIH->biBitCount==32)))
	{
	BMIH->biCompression = BI_BITFIELDS;
	redMask = (~(0xFFFFFFFF << redBits)) << (greenBits+blueBits);
	greenMask = (~(0xFFFFFFFF << greenBits)) << blueBits;
	blueMask = ~(0xFFFFFFFF << blueBits);
	return 1;
	}
  return 0;
  }

// file types supported:	24 and 32 bit true color images
//							16 and 32 bit bitfields compressed
//                 			8 bit colormap images
void BMPFile::WriteLineRGB (int y, unsigned char *buf)
  {
  if ((!image)||(!(flags&IMAGE_WRITE))) return;
  if (!bits) CompleteHeader();
  if (!bits) return;

  int x, w, h, pixelSize;

  pixelSize = (BMIH->biBitCount+7) >> 3;
  GetImageSize (w, h);
  if (BMIH->biHeight>=0) y = h-1-y;
  uint8 *tBuf = &bits[y*GetLineSize()];

  if ((BMIH->biBitCount==8) && (palette))
	for (x=0; x<w; x++)
	  tBuf[x] = palette->GetNearestPaletteIndex(TColor(buf[x*3],buf[x*3+1],buf[x*3+2]));

  else if ((BMIH->biCompression==BI_BITFIELDS)&&((BMIH->biBitCount==16)||(BMIH->biBitCount==32)))
	{
	int bm=blueMask, b=0, gm=greenMask, g=0, rm=redMask, r=0;
	while (bm) bm>>=1, b++;
	while (gm) gm>>=1, g++;
	while (rm) rm>>=1, r++;
	for (x=0; x<w; x++)
	  {
	  int p = (redMask&(buf[x*3]<<(r-8))) +
			  (greenMask&(buf[x*3+1]<<(g-8))) +
			  (blueMask&(buf[x*3+2]>>(8-b)));
	  if (pixelSize==2) ((uint16*)tBuf)[x] = p;
	  else ((uint32*)tBuf)[x] = p;
	  }
	}

  else if (BMIH->biBitCount>=24)
	for (x=0; x<w; x++)
	  {
	  tBuf[x*pixelSize+2] = buf[x*3];
	  tBuf[x*pixelSize+1] = buf[x*3+1];
	  tBuf[x*pixelSize] = buf[x*3+2];
	  }
  }

// file types supported:	4 and 8 bit colormap images
void BMPFile::WriteLineCLUT (int y, unsigned char *buf)
  {
  if ((!image)||(!(flags&IMAGE_WRITE))) return;
  if (!bits) CompleteHeader();
  if (!bits) return;

  int x, w, h, pixelSize;

  GetImageSize (w, h);
  if (BMIH->biHeight>=0) y = h-1-y;
  uint8 *tBuf = &bits[y*GetLineSize()];

  if (BMIH->biBitCount==4)
	for (x=0; x<w/2; x++)
	  tBuf[x] = (buf[2*x]<<4) + buf[2*x+1];

  else if (BMIH->biBitCount==8)
	for (x=0; x<w; x++)
	  tBuf[x] = buf[x];
  }

void BMPFile::WriteColorMap()
  {
  if ((!palette)||(!(flags&IMAGE_WRITE))||(!BMIH)) return;

  if ((BMIH->biBitCount > 8) && (BMIH->biClrUsed == 0)) return;
  unsigned int numColors = BMIH->biClrUsed;
  if (!numColors) numColors = 1 << BMIH->biBitCount;

  int i;
  RGBQUAD color;  color.rgbReserved=0;
  PALETTEENTRY pe;  pe.peRed=pe.peGreen=pe.peBlue=0;
  uint8 *pos = (uint8*)BMIH + BMIH->biSize;
  for (i=0; i<numColors; i++)
	{
	if (palette) palette->GetPaletteEntry(i, pe);
	color.rgbRed=pe.peRed; color.rgbGreen=pe.peGreen; color.rgbBlue=pe.peBlue;
	memcpy(pos, &color, sizeof(color));
	pos += 4;
	}
  }

void BMPFile::CompleteHeader()
  {
  BMIH->biClrUsed = 0;
  BMIH->biClrImportant = 0;

  if ((palette)&&(BMIH->biCompression!=BI_BITFIELDS))
	{
	BMIH->biClrUsed = palette->GetNumEntries();
	if ((BMIH->biBitCount<=8)&&(BMIH->biClrUsed==(1<<BMIH->biBitCount)))
	  BMIH->biClrUsed = 0;
	}

  int numColors = BMIH->biClrUsed;
  if ((!numColors)&&(BMIH->biBitCount <= 8)) numColors = 1<<BMIH->biBitCount;
  BMFH->bfOffBits = ((int)BMIH-(int)image) + BMIH->biSize + numColors*sizeof(RGBQUAD);
  if (BMIH->biCompression==BI_BITFIELDS) BMFH->bfOffBits += 12;
  BMIH->biSizeImage = 0;  //BMIH->biSizeImage = BMIH->biHeight * GetLineSize();

  int newsize = BMFH->bfOffBits + ABS(BMIH->biHeight)*GetLineSize();
  if (newsize > GetCurrentFileSize())
	if (!AdjustFileSize(newsize)) return;

  if (BMIH->biCompression==BI_BITFIELDS)		// BitFields Compression
	{
    uint32* mask = (uint32*)(&(((uint8*)BMIH)[BMIH->biSize]));
	mask[0] = redMask;
	mask[1] = greenMask;
	mask[2] = blueMask;
	}

  bits = &((uint8*)image)[BMFH->bfOffBits];
  WriteColorMap();
  }

int BMPFile::IsFileThisType(fstream *file)
  {
  BITMAPFILEHEADER fh;
  BITMAPINFOHEADER ih;
  file->read ((unsigned char*) &fh, sizeof(BITMAPFILEHEADER));
  file->read ((unsigned char*) &ih, sizeof(BITMAPINFOHEADER));
  file->clear();
  file->seekg(0);

  if (fh.bfType != 0x4D42) return 0;

  if (ih.biPlanes != 1) return -1;
  if (ih.biBitCount < 4) return -1;
  if ((ih.biCompression!=BI_RGB)&&(ih.biCompression!=BI_BITFIELDS)) return -1;

  return 1;
  }

